import pandas as pd
import os
from pathlib import Path
import sys
from datetime import datetime
from pytz import timezone
import numpy as np
import logging
from .dtgs import epoch2datetime, datetime2epoch, Dtg

logger = logging.getLogger(__name__)


class DataNotFoundError(Exception):
    pass


class DataFrameWithDtgMetadata(pd.DataFrame):
    """ Add DTG metadata to a pandas DataFrame """

    # See <https://pandas.pydata.org/pandas-docs/stable/development/
    #      extending.html#define-original-properties>
    # Define _internal_names and _internal_names_set for temporary properties
    # which WILL NOT be passed to manipulation results.
    # Define _metadata for normal properties which will be passed to
    # manipulation results.

    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ["temp_property"]
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ["_parent_dtg"]

    @property
    def _constructor(self):
        return DataFrameWithDtgMetadata


@np.vectorize
def mslp2sp_coeff(alt):
    """
    Coefficient to convert mean sea-level pressure to pressure at altitude alt
    """
    return 1.0 / (
        int(round(100000 / ((288 - 0.0065 * alt) / 288) ** 5.255)) / 100000
    )


def save_df_as_netatmo_csv(df, path, overwrite=False):
    path = Path(path).resolve()
    if path.exists() and not overwrite:
        raise FileExistsError("File {} exists".format(path))
    Path.mkdir(path.parent, parents=True, exist_ok=True)
    if pd.api.types.is_datetime64_any_dtype(df["time_utc"]):
        df["time_utc"] = datetime2epoch(df["time_utc"])
    elif not pd.api.types.is_integer_dtype(df["time_utc"]):
        raise ValueError("time_utc is nether an integer nor a datetime")
    if "pressure" in df.columns and "mslp" in df.columns:
        # In this case, we have recovered pressure from netatmo original file
        # when reading the df
        df = df.drop(["pressure"], axis=1)
    # It hurts to label pressures as mslp again, but this is needed
    # so that the csv files generated by this routine are consistent
    # in data type with the original ones provided by netatmo
    df = df.rename(columns=dict(mslp="pressure"))
    # Saving the adjusted df
    logger.info("Saving file %s", path)
    df.to_csv(path, index=None, mode="w")


def remove_str_prefix(string, prefix):
    if string.startswith(prefix):
        return string[len(prefix) :]
    else:
        return string


@np.vectorize
def shorten_stat_id(stat_id):
    new_id = remove_str_prefix(stat_id, prefix="enc:16:")
    return new_id[:8]


def read_netatmo_csv(
    path,
    dropna=True,
    fillna=dict(sum_rain_1=0.0),
    recover_pressure_from_mslp=False,
    drop_mslp=None,
):
    if drop_mslp is None:
        drop_mslp = recover_pressure_from_mslp
    if drop_mslp and not recover_pressure_from_mslp:
        raise ValueError(
            "Cannot have drop_mslp=True and recover_pressure_from_mslp=False"
        )

    data = pd.read_csv(path)
    # Replacing/removing invalid data (NaNs)
    if fillna:
        if not isinstance(fillna, dict):
            raise ValueError("The 'fillna' arg must be an instance of 'dict'")
        data = data.fillna(fillna)
    if dropna:
        data = data.dropna()

    # Drop the 'enc:16:' stat id prefix and shorten them to 8 chars
    data["id"] = shorten_stat_id(data["id"])

    # The netatmo data I got from Norway has mean sea-level pressure
    # instead of pressure, but the label is "pressure" there. Fix this.
    data = data.rename(columns={"pressure": "mslp"})
    pressure_cols = ["mslp"]
    if recover_pressure_from_mslp:
        data["pressure"] = mslp2sp_coeff(data["alt"]) * data["mslp"]
        pressure_cols += ["pressure"]
    # Make sure columns 'id', 'time_utc', 'lat' and 'lon' appear
    # first AND in this order
    data = data[
        ["id", "time_utc", "lat", "lon", "alt"]
        + pressure_cols
        + ["temperature", "humidity", "sum_rain_1"]
    ]
    if drop_mslp:
        data = data.drop(["mslp"], axis=1)

    data["time_utc"] = epoch2datetime(data["time_utc"])

    return data.sort_values(by="time_utc").reset_index(drop=True)


def gen_netatmo_fpaths_for_dates(dates, rootdir):
    """
    Returns an generator for the paths of the NetAtmo CSV files that need
    to be read in order to get data for date=dt

    NetAtmo CSV files are assumed to have full paths following the pattern
            "rootdir/%Y%m/%d/%Y%m%dT%H%M%SZ.csv"
    """
    try:
        date_iter = iter(dates)
    except (TypeError):
        date_iter = [dates]
    rootdir = Path(rootdir).resolve()
    for date in sorted(date_iter):
        yyyymm = date.strftime("%Y%m")
        dd = date.strftime("%d")
        for f in sorted((rootdir / yyyymm / dd).glob("*.csv")):
            yield f


def read_netatmo_data_for_dates(
    dates, rootdir, recover_pressure_from_mslp=False, drop_mslp=None,
):
    """
    Returns a pd dataframe with data from NetAtmo files for given dates
    """
    data_from_all_files = {}
    for f in gen_netatmo_fpaths_for_dates(dates, rootdir):
        data_from_all_files[f.stem] = read_netatmo_csv(
            f,
            drop_mslp=drop_mslp,
            recover_pressure_from_mslp=recover_pressure_from_mslp,
        )
    if len(data_from_all_files) > 0:
        return pd.concat(data_from_all_files, ignore_index=True)
    else:
        raise DataNotFoundError(
            "Could not find data for date(s)={} under {}".format(
                dates, rootdir
            )
        )


def datetime_of_netatmo_file(f, fmt="%Y%m%dT%H%M%SZ.csv"):
    dt = datetime.strptime(f.name, fmt)
    return dt.replace(tzinfo=timezone("utc"))


# Reading data for a given DTG instead of date
def gen_netatmo_files_for_time_window(start, end, rootdir):
    """
    List of NetAtmo files corresponding to datetimes within [start, end]

    Returns a list of paths leading to NetAtmo files whose names correspond
    to timestamps lying within the closed interval [start, end].
    """
    start, end = sorted((start, end))
    dates = pd.date_range(start.date(), end.date(), freq="1d")
    for f in gen_netatmo_fpaths_for_dates(dates, rootdir):
        t = datetime_of_netatmo_file(f)
        if t >= start and t <= end:
            yield f


def gen_netatmo_fpaths_for_dtg(dtg, rootdir):
    """
    Paths to NetAtmo files that are likely to contain data for DTG=dtg
    """
    # Add a few minutes to the assimilation window, as
    # netatmo data is gathered every ~10 minutes
    start = dtg.cycle_start - pd.Timedelta("15 minutes")
    end = dtg.cycle_end + pd.Timedelta("15 minutes")
    return gen_netatmo_files_for_time_window(start, end, rootdir=rootdir)


def rm_moving_stations(df):
    # Remove inconsistent stations
    df_grouped_by_id = df.groupby(["id"], as_index=False, sort=False)
    df_temp = df_grouped_by_id[["lat", "lon", "alt"]].agg(["max", "min"])
    inconsistent_stations = df_temp[
        (abs(df_temp["lat"]["max"] - df_temp["lat"]["min"]) >= 1e-4)
        | (abs(df_temp["lon"]["max"] - df_temp["lon"]["min"]) >= 1e-4)
        | (abs(df_temp["alt"]["max"] - df_temp["alt"]["min"]) >= 1e-2)
    ].index.values
    df_ok = df[~df["id"].isin(inconsistent_stations)]
    df_inconsistent = df[df["id"].isin(inconsistent_stations)]
    return df_ok, df_inconsistent


def remove_duplicates_within_cycle(df, dtg):
    """
    Remove duplicates for each station by keeping the one closest to DTG

    If a station reports data multiple times within the time window,
    then keep the one closest to the passed DTG and discard the rest.
    """
    # It seems that using groupby followed by "apply" with a custom
    # function is about 20x slower.
    # Sort using df.loc is better than adding a new col to sort by it. See
    # <stackoverflow.com/questions/39525928/pandas-sort-lambda-function>
    df = df.loc[(abs(df["time_utc"] - dtg)).sort_values().index]
    # We don't need sorting with groupby. HDBSCAN clustering, however,
    # seems to be sensitive to the order in which the rows appear (but
    # not very much so). This seems to be knwon, see, e.g.,
    # <https://github.com/scikit-learn-contrib/hdbscan/issues/265>
    df = df.groupby(["id"], as_index=False, sort=False).first()
    return df


def read_netatmo_data_for_dtg(
    dtg,
    rootdir,
    return_list_of_removed_stations=False,
    recover_pressure_from_mslp=False,
    drop_mslp=None,
):
    if not isinstance(dtg, Dtg):
        dtg = Dtg(dtg)
    data_from_all_files = {}
    for f in gen_netatmo_fpaths_for_dtg(dtg, rootdir=rootdir):
        data_from_all_files[f.stem] = read_netatmo_csv(
            f,
            drop_mslp=drop_mslp,
            recover_pressure_from_mslp=recover_pressure_from_mslp,
        )
    if len(data_from_all_files) > 0:
        df = pd.concat(data_from_all_files, ignore_index=True)
        # Remove DTGs that do not belong to the assimilation time window
        if dtg.assimilation_window.closed_left:
            mask_left = df["time_utc"] >= dtg.assimilation_window.left
        else:
            mask_left = df["time_utc"] > dtg.assimilation_window.left
        if dtg.assimilation_window.closed_right:
            mask_right = df["time_utc"] <= dtg.assimilation_window.right
        else:
            mask_right = df["time_utc"] < dtg.assimilation_window.right
        df = df[mask_left & mask_right]

        # Remove stations that change (lat, lon, alt) within cycle
        df, df_inconsistent_stations = rm_moving_stations(df)
        if len(df_inconsistent_stations.index) > 0:
            logger.warn(
                'Removed %d obs from %d "moving" stations in DTG=%s',
                len(df_inconsistent_stations.index),
                len(df_inconsistent_stations["id"].unique()),
                dtg,
            )
        # Remove duplicate data for stations within cycle by keeping
        # the data reported for the timestamp closest to the passed DTG
        df = remove_duplicates_within_cycle(df, dtg)
    else:
        raise DataNotFoundError(
            "Could not find data for DTG={} under {}".format(dtg, rootdir)
        )

    # Add DTG metadata, to a future reference of how the data was collated
    df = DataFrameWithDtgMetadata(df)
    df._parent_dtg = dtg

    logger.debug(
        "Dataframe has %d observations and occupies %.3f MB of memory",
        len(df.index),
        df.memory_usage(deep=True).sum() / (1024.0 ** 2),
    )

    if return_list_of_removed_stations:
        return df, df_inconsistent_stations
    else:
        return df



import argparse
from datetime import datetime
import subprocess
import os
from pathlib import Path
from netatmoqc.commands_functions import (
    cluster_obs_single_dtg,
    select_stations,
)


def get_parsed_args(program_name):
    # Define main parser and general options
    parser = argparse.ArgumentParser(
        prog=program_name,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    try:
        fpath = Path(os.getenv("NETATMOQC_CONFIG_PATH", "config.toml"))
        default_conf_path = fpath.resolve(strict=True)
    except (FileNotFoundError):
        default_conf_path = (
            Path(os.getenv("HOME")) / ".netatmoqc" / "config.toml"
        )
    parser.add_argument(
        "-config_file",
        metavar="CONFIG_FILE_PATH",
        default=default_conf_path,
        type=Path,
        help=(
            "Path to the config file. The default is whichever of the "
            + "following is first encountered: "
            + "(i) The value of the 'NETATMOQC_CONFIG_PATH' envvar or "
            + "(ii) './config.toml'. If both (i) and (ii) are missing, "
            + "then the default will become "
            + "'"
            + str(Path("$HOME/.netatmoqc/config.toml"))
            + "'"
        ),
    )
    parser.add_argument(
        "-loglevel",
        default="info",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
        help="What type of info should be printed to the log",
    )
    parser.add_argument(
        "--mpi",
        action="store_true",
        help=(
            "Enable MPI parallelisation in some parts of the code. "
            + "Requires that the code be installed with support to MPI."
        ),
    )
    parser.add_argument(
        "--show",
        action="store_true",
        help="Open a browser window to show results on a map",
    )
    parser.add_argument(
        "--savefig",
        action="store_true",
        help="Saves an html file showing results on a map",
    )

    # Configure the main parser to handle the commands
    # From python v3.7, add_subparsers accepts a 'required' arg. We
    # could use required=True in the definition below, but, since we
    # want to be able to use python >= 3.6.10, we won't. The requirement
    # of having a command passed will be enforced in the main.py file.
    subparsers = parser.add_subparsers(
        title="commands",
        dest="command",
        description=(
            "Valid commands for {0} (note that commands also accept their "
            + "own arguments, in particular [-h]):"
        ).format(program_name),
        help="command description",
    )

    # Configure parser for the "cluster" command
    parser_cluster = subparsers.add_parser(
        "cluster", help="Run cluster on NetAtmo obs for a single DTG"
    )
    parser_cluster.add_argument(
        "-dtg",
        type=lambda s: datetime.strptime(s, "%Y%m%d%H"),
        default=None,
        help="""
        If not passed, then the first DTG in the config file will be used.
        """,
    )
    parser_cluster.set_defaults(func=cluster_obs_single_dtg)

    # Configure parser for the "select" command
    parser_select = subparsers.add_parser(
        "select",
        help="Select NetAtmo stations via clustering over a range of DTGs",
    )
    parser_select.set_defaults(func=select_stations)

    # Configure parser for the "apps" command
    parser_apps = subparsers.add_parser(
        "apps", help="Start a server for the provided Dash apps"
    )
    parser_apps.set_defaults(
        func=lambda args: subprocess.run(
            [
                "sh",
                Path(__file__).parent.resolve() / "bin" / "start_apps",
                args.config_file.resolve(),
            ]
        )
    )

    args = parser.parse_args()

    return args



#!/usr/bin/env python3
import logging

logger = logging.getLogger(__name__)
import pandas as pd
from pandas.tseries.frequencies import to_offset
from datetime import datetime, timedelta
from pytz import timezone
import numpy as np
from functools import partial


@np.vectorize
def epoch2datetime(epoch):
    return datetime.fromtimestamp(epoch, timezone("utc"))


@np.vectorize
def datetime2epoch(dt):
    reference = datetime(1970, 1, 1, tzinfo=timezone("utc"))
    epoch = int((dt - reference).total_seconds())
    return epoch


class Dtg(datetime):
    def __init__(self, *args, **kwargs):
        # Set a default of 3H for the length of a cycle.
        # Without the context provided by a cycle length, a
        # DTG has no meaning.
        self.set_cycle_length(kwargs.get("cycle_length", "3H"))

    def __new__(cls, *args, **kwargs):
        # cycle_length checks will only be performed in __init__
        kwargs.pop("cycle_length", None)

        # timezone will be set to UTC if not passed
        try:
            tzinfo = args[7]
            args = args[:7]
        except (IndexError):
            tzinfo = kwargs.pop("tzinfo", timezone("utc"))
        if isinstance(tzinfo, str):
            tzinfo = timezone(tzinfo)
        if "tzinfo" not in kwargs:
            kwargs["tzinfo"] = tzinfo

        # Make it possible to create instance using either string or an
        # instance of datetime as the 1st positional argument. Keep support
        # for the datetime instantiation behaviour (with integers for year,
        # month, day, etc).
        if len(args) > 0 and not isinstance(args[0], int):
            if not any(isinstance(args[0], c) for c in [str, datetime]):
                raise TypeError(
                    "Cannot convert input '{}' ".format(args[0])
                    + "of type '{}' ".format(type(args[0]).__name__)
                    + "to {}".format(cls.__name__)
                )
            elif len(args) > 1:
                raise ValueError(
                    "Only one positional argument is allowed "
                    + "if the first argument is not an integer"
                )

            try:
                # Add support to the YYYYMMDDHH standard DTG format
                # Assumed to be in UTC unless tzinfo is specified
                dt = datetime.strptime(args[0], "%Y%m%d%H")
                dt = dt.replace(tzinfo=kwargs["tzinfo"])
            except (ValueError, TypeError):
                # pd.Timestamp is flexible with string and datetime-like args
                # It is more convenient using pd.Timestamp than trying to
                # guess the type and format of args[0] manually
                ts_kwargs = kwargs.copy()
                try:
                    if args[0].tzinfo is not None:
                        ts_kwargs.pop("tzinfo")
                except (AttributeError):
                    pass
                dt = pd.Timestamp(*args, **ts_kwargs).to_pydatetime()
                dt = dt.astimezone(kwargs["tzinfo"])

            # Clear args and pass only kwargs to datetime's __new__
            args = ()
            for att in [
                "year",
                "month",
                "day",
                "hour",
                "minute",
                "second",
                "microsecond",
                "tzinfo",
            ]:
                kwargs[att] = getattr(dt, att)

        new_instance = super().__new__(cls, *args, **kwargs)
        # Make sure DTGs don't have subdivisions smaller than one hour
        # This can easily happen if the instance is created based on string
        # or datetime-like arguments, or when it is created as a result of an
        # addition/subtraction operation
        eqv_valid_instance = super().__new__(
            cls,
            new_instance.year,
            new_instance.month,
            new_instance.day,
            new_instance.hour,
            tzinfo=new_instance.tzinfo,
        )
        if new_instance.as_datetime() != eqv_valid_instance.as_datetime():
            msg = "Dtg {} not allowed: ".format(new_instance.as_datetime())
            msg += "Minimum allowed Dtg subdivision is an hour"
            raise ValueError(msg)

        return new_instance

    def __reduce_ex__(self, protocol):
        # We define this method so that pickle knows how to recover a pickled
        # object of this class. If we don't do this, then pickle tries to pass
        # a bytes object to the classe's __new__ method upon inpickling. This
        # also indirectly affects copy and deepcopy. See a similar issue at
        # <https://stackoverflow.com/questions/59571625/
        #  unable-to-pickle-datetime-subclass>
        # Using functools.partial to be able to pass kwargs
        callable_for_new = partial(type(self), cycle_length=self.cycle_length)
        args = (self.isoformat(),)
        return callable_for_new, args

    # Handy type conversions
    def as_pandas_Timestamp(self):
        return pd.Timestamp(self)

    def as_datetime(self):
        return self.as_pandas_Timestamp().to_pydatetime()

    # Useful little methods/properties
    @property
    def cycle_length(self):
        return self._cycle_length

    @property
    def next(self):
        return self + self.cycle_length

    @property
    def prev(self):
        return self - self.cycle_length

    @property
    def cycle_start(self):
        return self.as_pandas_Timestamp() - 0.5 * self.cycle_length.delta

    @property
    def cycle_end(self):
        return self.as_pandas_Timestamp() + 0.5 * self.cycle_length.delta

    @property
    def assimilation_window(self):
        return pd.Interval(self.cycle_start, self.cycle_end, closed="left")

    # cycle_length-related methods
    def compatible_with_cycle_length(self, cycle_length="self"):
        if cycle_length is None:
            return True
        elif cycle_length == "self":
            cycle_length = self.cycle_length
        else:
            cycle_length = to_offset(cycle_length)
        # It is convenient to check this using pandas Timestamp objects
        # At the time of writing this, inheriting from pd.Timestamp lead
        # to strange behaviours.
        dtg = pd.Timestamp(self)
        return (dtg - dtg.normalize()) % cycle_length == timedelta(0)

    def set_cycle_length(self, cycle_length):
        # Set a default of 3H for the length of a cycle.
        # Without the context provided by a cycle length, a
        # DTG has no meaning.
        cycle_length = to_offset(cycle_length)
        # Validate cycle_length
        min_cycle_length = to_offset("1H")
        ref_date = datetime(2000, 1, 1)
        new_date = ref_date + cycle_length
        min_allowed_new_date = ref_date + min_cycle_length
        if new_date < min_allowed_new_date:
            raise ValueError(
                "Min allowed cycle_length is "
                + str(min_cycle_length)
                + ". Passed cycle_length={}".format(cycle_length)
            )
        if not self.compatible_with_cycle_length(cycle_length):
            raise ValueError(
                "Dtg {} not compatible with cycle_length {}".format(
                    self, cycle_length
                )
            )
        self._cycle_length = cycle_length

    # Override some methods to ensure results are consistent with Dtg type
    def replace(self, *args, **kwargs):
        new = super().replace(*args, **kwargs)
        return Dtg(new, cycle_length=self.cycle_length)

    # Behaviour under action of + and - operands
    def __add__(self, delta):
        new_datetime = self.as_datetime() + delta
        return self.__class__(new_datetime, cycle_length=self.cycle_length)

    def __radd__(self, delta):
        return self.__add__(delta)

    def __sub__(self, other):
        new = self.as_datetime() - other
        try:
            # Casting to Dtg will work provided that "other" is a
            # timedelta-like obj and it does not violate cycle_length
            return self.__class__(new, cycle_length=self.cycle_length)
        except (TypeError):
            # This will happen if "other" is any datetime-like object
            return new

    # Make sure we consider cycle_length when comparing for equality
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        eq_timestamp = super().__eq__(other)
        same_cycle_length = self.cycle_length == other.cycle_length
        return eq_timestamp and same_cycle_length

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.strftime("%Y-%m-%dT%H %Z")

    def __repr__(self):
        return "{}('{}', cycle_length='{}')".format(
            self.__class__.__name__,
            self.isoformat(),
            self.cycle_length.freqstr,
        )


class DtgContainer:
    def __init__(self, data=None, start=None, end=None, cycle_length=None):
        """Container for instances of the Dtg class"""

        # Validate input:
        # (i) data and (start, end) are mutually exclusive (avoid ambiguities)
        if data is None:
            if (start is None) or (end is None):
                raise ValueError(
                    "Need 'start' and 'end' and if 'data' is not passed"
                )
            # msg_which_input and input_vals will be used later on in the
            # cycle_length validation
            msg_which_input = "'start' and 'end'"
            input_vals = [start, end]
        else:
            if (start is not None) or (end is not None):
                raise ValueError(
                    "Pass either just 'data' or both 'start' and 'end'"
                )
            msg_which_input = "the elements of 'data'"
            input_vals = data
        # (ii) Decide whether to get cycle_length from
        #      data/start/end or from class arg
        try:
            # Calling method ".freqstr" because, since pandas v1.1.0, "freq"
            # objects stopped being hashabe and the following line would cause
            # "TypeError: unhashable type: 'pandas._libs.tslibs.offsets.Hour'"
            data_c_length = list(
                set(d.cycle_length.freqstr for d in input_vals)
            )
        except (AttributeError):
            data_c_length = []

        if cycle_length is None:
            if len(data_c_length) == 1:
                cycle_length = data_c_length[0]
            else:
                raise ValueError(
                    "No (or inconsistent) 'cycle_length' info found in "
                    + "{}, and no 'cycle_length' ".format(msg_which_input)
                    + "passed to '{}'".format(self.__class__.__name__)
                )
        elif len(data_c_length) > 0:
            logger.warning(
                "The value(s) of the 'cycle_length' attribute(s) of "
                + "%s will be replaced by the 'cycle_length' value "
                + "passed to the '%s' constructor (=%s)",
                msg_which_input,
                self.__class__.__name__,
                cycle_length,
            )

        # Populate attributes
        self._cycle_length = to_offset(cycle_length)
        if data is None:
            self._start = Dtg(start, cycle_length=self.cycle_length)
            self._end = Dtg(end, cycle_length=self.cycle_length)
            self._data = None
        else:
            self._start = None
            self._end = None
            # Convert data to Dtg for consistency, and make it immutable
            self._data = tuple(
                Dtg(d, cycle_length=self.cycle_length) for d in data
            )

    @property
    def cycle_length(self):
        return self._cycle_length

    def __len__(self):
        if self._data is None:
            max_index = abs((self._end - self._start).total_seconds())
            max_index /= self.cycle_length.nanos * 1e-9
            return int(max_index) + 1
        else:
            return len(self._data)

    def __getitem__(self, item):
        # We don't need to define __iter__ when we make an obj indexable
        if self._data is not None:
            return self._data[item]
        else:
            sign = np.sign((self._end - self._start).total_seconds())

            def calc_nth_item(n):
                if (n > len(self) - 1) or (n < -len(self)):
                    raise IndexError(
                        "{} index out of range".format(self.__class__.__name__)
                    )
                return self._start + sign * (n % len(self)) * self.cycle_length

            if isinstance(item, slice):
                return [calc_nth_item(i) for i in range(len(self))[item]]
            else:
                return calc_nth_item(item)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return (
            other.cycle_length == self.cycle_length
            and other._data == self._data
            and other._start == self._start
            and other._end == self._end
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        rtn = "{} object:\n".format(self.__class__.__name__)
        if self._data is None:
            rtn += "    start={}, end={}\n".format(self._start, self._end)
        else:
            rtn += "    data=[{}]\n".format(", ".join(map(str, self._data)))
        rtn += "    cycle_length={}".format(self.cycle_length)
        return rtn

    __str__ = __repr__



import logging
import argparse
import subprocess
import multiprocessing
from joblib import Parallel, delayed, parallel_backend
import time
import os
import sys
import psutil
import numpy as np
import pandas as pd
from datetime import datetime
from pathlib import Path
from functools import partial
from netatmoqc.config_parser import read_config
from netatmoqc.dtgs import Dtg
from netatmoqc.logs import get_logger
from netatmoqc.mpi import mpi_parallel
from netatmoqc.plots import make_dbscan_fig
from netatmoqc.load_data import rm_moving_stations, DataNotFoundError
from netatmoqc.clustering import cluster_netatmo_obs, report_clustering_results

logger = logging.getLogger(__name__)

CYAN = "\u001b[36m"
RESET_COLOUR = "\u001b[0m"


def cluster_obs_single_dtg(args):
    """ Implements the "cluster" command """

    try:
        dtg = Dtg(args.dtg)
    except (TypeError):
        dtg = None
    config = read_config(args.config_file)
    if dtg is None:
        dtg = config.dtgs[0]
    logger.info(
        "Running %s on obs for " + CYAN + "DTG=%s" + RESET_COLOUR + "...",
        config.clustering_method,
        dtg,
    )
    df = cluster_netatmo_obs(
        config, dtg=args.dtg, sort_by_cluster_size=(args.show or args.savefig)
    )
    report_clustering_results(df)
    if args.show or args.savefig:
        logger.info("Preparing figure...")
        fig = make_dbscan_fig(df)
        if args.savefig:
            fig.write_html("cluster_obs_single_dtg.html")
        if args.show:
            fig.show()
    logger.info("Done.")


def _select_stations_single_dtg(dtg, config, args):
    """ This is what "select_stations" does for each DTG """

    logger = get_logger(__name__, args.loglevel)
    logger.info(CYAN + "DTG=%s" + RESET_COLOUR + ": Started", dtg)

    # Some control of oversubscription if using mpi.
    # This is not needed when using single-host joblib with "loky"
    cpu_share = -1
    if args.mpi:
        proc_parent = psutil.Process(os.getppid())
        proc_family_size = len(proc_parent.children())
        cpu_share = multiprocessing.cpu_count() // proc_family_size

    tstart = time.time()
    try:
        df, rmvd_stations = cluster_netatmo_obs(
            config,
            dtg,
            return_list_of_removed_stations=True,
            num_threads=cpu_share,
        )
    except DataNotFoundError as e:
        logger.warning(e)
        return pd.DataFrame(), pd.DataFrame(), pd.DataFrame()

    selected_cols = ["id", "lat", "lon", "alt"]
    rejected_stat_mask = df["cluster_label"] < 0
    rejected_stations = df[selected_cols][rejected_stat_mask]
    accepted_stations = df[selected_cols][~rejected_stat_mask]
    moving_stations = rmvd_stations[selected_cols]

    logger.info("DTG=%s: Done. Elapsed: %.1fs", dtg, time.time() - tstart)
    return accepted_stations, rejected_stations, moving_stations


def select_stations(args):
    """ Implements the "select" command

    This function calls "_select_stations_single_dtg" for each DTG
    (in parallel if requested/possible), and then processes, gathers
    and saves the results.
    """

    tstart_selection = time.time()
    config = read_config(args.config_file)

    # Process DTGs in parallel if possible/requested
    if args.mpi:
        logger.info("Parallelising tasks over DTGs using MPI")
        func = partial(_select_stations_single_dtg, config=config, args=args)
        results = mpi_parallel(func, config.dtgs)
    else:
        logger.info("Parallelising tasks over DTGs within a single host")
        # Using the "loky" backend helps avoiding oversubscription
        # of cpus in child processes
        n_procs = int(os.getenv("NETATMOQC_MAX_PYTHON_PROCS", 1))
        with parallel_backend("loky"):
            results = Parallel(n_jobs=n_procs)(
                delayed(_select_stations_single_dtg)(dtg, config, args)
                for dtg in config.dtgs
            )
    logger.info(
        "End of survey over %d DTGs. Combining results..." + "\n",
        len(config.dtgs),
    )
    # Gather results: Only accepted and rejected stations for now. Moving
    # stations will be handled after post-processing accepted and rejected
    accepted_stations = pd.concat((r[0] for r in results), ignore_index=True)
    rejected_stations = pd.concat((r[1] for r in results), ignore_index=True)

    accepted_stations, moving_accepted_stations = rm_moving_stations(
        accepted_stations
    )
    if len(moving_accepted_stations.index) > 0:
        logger.warning(
            'Removed %d "moving" stations from the list of accepted stations',
            len(moving_accepted_stations["id"].unique()),
        )

    rejected_stations, moving_rejected_stations = rm_moving_stations(
        rejected_stations
    )
    if len(moving_rejected_stations.index) > 0:
        logger.warning(
            'Removed %d "moving" stations from the list of removed stations',
            len(moving_rejected_stations["id"].unique()),
        )

    # Let's recover some stations for the list of rejected if they
    # also appear among the accepted_stations. Do this based on a
    # tolerance criteria.
    logger.info(
        "Applying station_rejection_tol=%.2f to rescue "
        + 'stations from the "rejected" list',
        config.commands.select.station_rejection_tol,
    )
    accept_counts_per_id = accepted_stations["id"].value_counts()
    reject_counts_per_id = rejected_stations["id"].value_counts()
    rescued_rejected_stats = []
    for stat_id in reject_counts_per_id.index:
        reject_count = reject_counts_per_id[stat_id]
        try:
            accept_count = accept_counts_per_id[stat_id]
        except (KeyError):
            accept_count = 0
        rejection_rate = 1.0 * reject_count / (reject_count + accept_count)
        if rejection_rate < config.commands.select.station_rejection_tol:
            rescued_rejected_stats.append(stat_id)
    # Using a rescued_rejected_stats list to modify the dataframes
    # accepted_stations/rejected_stations is much faster than modifying
    # the dataframes inside the loop above
    rescued_rows = rejected_stations.loc[
        rejected_stations["id"].isin(rescued_rejected_stats), :
    ]
    accepted_stations = accepted_stations.append(
        rescued_rows, ignore_index=True
    )
    rejected_stations = rejected_stations.drop(rescued_rows.index)
    if len(rescued_rejected_stats) > 0:
        logger.info(
            "    > Rescuing %d stations rejected in less than %.1f%% of "
            + "occurences",
            len(rescued_rejected_stats),
            100 * config.commands.select.station_rejection_tol,
        )
    else:
        logger.info('    > No stations recovered from "rejected"')

    # We don't need duplicate entries
    accepted_stations = accepted_stations.drop_duplicates(subset=["id"])
    rejected_stations = rejected_stations.drop_duplicates(subset=["id"])

    # Now we can handle moving stations
    moving_stations = pd.concat(
        [r[2] for r in results]
        + [moving_accepted_stations, moving_rejected_stations,],
        ignore_index=True,
    )
    unique_moving_stations = moving_stations["id"].unique()

    rejected_stations = rejected_stations[
        ~rejected_stations["id"].isin(unique_moving_stations)
    ]
    accepted_stations = accepted_stations[
        ~(
            accepted_stations["id"].isin(unique_moving_stations)
            | accepted_stations["id"].isin(rejected_stations["id"])
        )
    ]
    if len(unique_moving_stations) > 0:
        logger.warning(
            "\n A total of %d unique stations were removed from both\n"
            + " accepted and rejected lists because they changed\n"
            + " at least one of (lat, lon, alt) throughout the survey\n"
            + " These will be saved separately in the moving_stations.csv file",
            len(unique_moving_stations),
        )

    total_nstations = sum(
        map(
            len,
            [
                accepted_stations.index,
                rejected_stations.index,
                unique_moving_stations,
            ],
        )
    )
    n_selected = len(accepted_stations.index)
    ratio = 100.0 * n_selected / total_nstations

    elapsed = time.time() - tstart_selection
    logger.info(
        "Station selection over %d DTGs performed in %.1fs (~%.1fs per DTG)"
        + "\n",
        len(config.dtgs),
        elapsed,
        elapsed / len(config.dtgs),
    )

    logger.info(
        "%d stations (%.2f%% out of %d) "
        + "were consistently accepted throughout the survey",
        n_selected,
        ratio,
        total_nstations,
    )
    logger.info(
        "That's %d rejections (%.2f%% rejection rate)",
        total_nstations - n_selected,
        100.0 - ratio,
    )

    # Saving results
    fpath = Path().resolve() / "accepted_stations.csv"
    logger.info("Saving accepted stations data to %s", fpath)
    accepted_stations.to_csv(fpath, index=None, mode="w")

    fpath = Path().resolve() / "rejected_stations.csv"
    logger.info("Saving rejected station data to %s", fpath)
    rejected_stations.to_csv(fpath, index=None, mode="w")

    fpath = Path().resolve() / "moving_stations.csv"
    logger.info("Saving moving stations data to %s", fpath)
    moving_stations.to_csv(fpath, index=None, mode="w")

    if args.show or args.savefig:
        accepted_stations["cluster_label"] = np.zeros(
            len(accepted_stations.index)
        )
        rejected_stations["cluster_label"] = -np.ones(
            len(rejected_stations.index)
        )

        if len(moving_stations.index) > 0:
            statid2label_dict = dict(
                (statid, -(i + 12))
                for i, statid in enumerate(unique_moving_stations)
            )

            @np.vectorize
            def statid2label(sid):
                return statid2label_dict[sid]

            moving_stations["cluster_label"] = statid2label(
                moving_stations["id"]
            )

        df = pd.concat(
            [rejected_stations, moving_stations, accepted_stations],
            ignore_index=True,
        )
        logger.info("Preparing figure...")
        fig = make_dbscan_fig(df)
        if args.savefig:
            fig.write_html("select_stations.html")
        if args.show:
            fig.show()



#!/usr/bin/env python3
import numpy as np
import pandas as pd
from hdbscan import HDBSCAN, RobustSingleLinkage
from sklearn.cluster import DBSCAN, OPTICS
from sklearn import metrics
from sklearn.neighbors import LocalOutlierFactor
from numba import njit, prange
import time
import sys
import logging
import io
from functools import wraps

# Local imports
from .config_parser import UndefinedConfigValue
from .load_data import read_netatmo_data_for_dtg
from .metrics import calc_distance_matrix

logger = logging.getLogger(__name__)


def suspendlogging(func):
    """
    Adapted from: <https://stackoverflow.com/questions/7341064/
                   disable-logging-per-method-function>
    """

    @wraps(func)
    def inner(*args, **kwargs):
        previousloglevel = logger.getEffectiveLevel()
        try:
            logger.setLevel(logging.WARN)
            return func(*args, **kwargs)
        finally:
            logger.setLevel(previousloglevel)

    return inner


def get_weights_array(
    df,
    pairwise_diff_weights={},
    skip=["id", "time_utc"],
    default=1,
    default_geodist=1,
):
    """
    Takes a pandas dataframe and a {column_name:weight} dictionary
    and returns a numpy array of weights to be passed to the routine
    that calculates the distance matrix.

    The "skip" arg lists columns that will not enter the clustering
    and should therefore be skipped.

    If the weight for a non-skipped column of the input dataframe
    are not defined in pairwise_diff_weights, then it will be set to
    default_weight, with a warning.

    Columns "lat" and "lon" are treated specially, in that they are
    not assigned a weight individually, but rather a single weight
    should be assigned to the "geo_dist" property.
    """
    weights = []
    weights_already_parsed = []
    for col in df.columns:
        if col in skip:
            continue
        weight_dict_key = col
        if col in ["lat", "lon"]:
            weight_dict_key = "geo_dist"
        if weight_dict_key in weights_already_parsed:
            continue
        weights_already_parsed.append(weight_dict_key)
        try:
            weight = pairwise_diff_weights[weight_dict_key]
            if weight is UndefinedConfigValue:
                raise KeyError
            weights.append(weight)
        except (KeyError):
            # If the user has not explicitely set the weight for this property
            if weight_dict_key == "geo_dist":
                weights.append(default_geodist)
            else:
                weights.append(default)
    return np.array(weights, dtype=np.float64)


def get_silhouette_samples(df, distance_matrix):
    # We will only consider true clusters for the calculation of
    # the silhouette coeff, i.e., points with label<0 will be excluded from
    # the calculation. This means that we need to use a slice of the distance
    # matrix in the calculation corresponding to the valid points.
    i_valid_obs = df.reset_index(drop=True).index[df["cluster_label"] >= 0]
    if len(i_valid_obs) == 0:
        logger.warn("No valid obs to calculate silhouette coeffs with")
        return -np.ones(len(df.index))

    reduced_labels = df["cluster_label"].iloc[i_valid_obs]
    if len(reduced_labels.unique()) == 1:
        logger.warn("Only one cluster to calculate silhouette coeffs with")
        return np.ones(len(df.index))

    reduced_dist_matrix = distance_matrix[i_valid_obs, :][:, i_valid_obs]
    # Using an intermediate np array is about 10x faster than
    # using df.at[iloc, 'silhouette_score'] = coeff
    all_silhouette_scores = np.empty(len(df.index))
    all_silhouette_scores[:] = np.nan

    reduced_silhouette_scores = metrics.silhouette_samples(
        X=reduced_dist_matrix, labels=reduced_labels, metric="precomputed"
    )
    for iloc, coeff in zip(i_valid_obs, reduced_silhouette_scores):
        all_silhouette_scores[iloc] = coeff

    return all_silhouette_scores


@njit(cache=True)
def get_rm_obs_label_offset(labels):
    """
    Integer value to be added to the cluster labels of outliers

    Constant integer value to be added to the cluster labels of
    the observations found to be outliers by our refinement schemes

    "labels" must be a numpy array for the numba jit compilation to work
    """
    rm_obs_label_offset = 10
    unique_labels = np.unique(labels)
    if unique_labels[0] < -1:
        max_abs_label = abs(unique_labels[0])
    else:
        max_abs_label = max(unique_labels[-1], 0)

    while max_abs_label % rm_obs_label_offset != max_abs_label:
        rm_obs_label_offset *= 10
    if unique_labels[0] < -1:
        rm_obs_label_offset //= 10

    return rm_obs_label_offset


# Routines related to the "iterative" outlier removal method
@njit(cache=True)
def _trim_np_array(array, perc, min_size=3):
    """Removes extreme valus from both sides of an array

    Takes an array and removes the "50*perc %" largest and
    "50*perc %" smallest elements, provided that the resulting
    trimmed array has a length no smaller tyan min_size
    """
    len_array = len(array)
    n_rm_each_side = int(abs(perc) * 0.5 * len_array)
    truncated_size = len_array - 2 * n_rm_each_side
    if (truncated_size != len_array) and (truncated_size >= min_size):
        array.sort()
        array = array[n_rm_each_side:-n_rm_each_side]
    return array


@njit(cache=True, parallel=True)
def _truncated_means_and_stds(df, perc, min_size=3):
    """
    Truncated means and stdevs for all columns in the dataframe df

    Takes a numpy version "df" of a pandas dataframe (obtained by
    calling the ".to_numpy" method on the original dataframe) and
    returns the truncated mean values and standard deviations for
    all columns. See the _trim_np_array routine for a description
    of the perc and min_size args.
    """
    ncols = df.shape[1]
    means = np.zeros(ncols)
    stds = np.zeros(ncols)
    for icol in prange(ncols):
        array = _trim_np_array(df[:, icol], perc, min_size)
        means[icol] = np.mean(array)
        stds[icol] = np.std(array)
    return means, stds


@njit(cache=True, parallel=True)
def _get_new_labels_for_deviating_obs_one_iter(
    df, max_n_stdev_around_mean, rm_obs_label_offset, truncate, weights
):
    """
    A single iteration of the _get_new_labels_for_deviating_obs function below
    """
    nrows, ncols = df.shape
    unique_labels = np.unique(df[:, -1])

    for label in unique_labels[unique_labels > -1]:
        indices_for_label = np.transpose(np.argwhere(df[:, -1] == label))[0]

        cols_taken_into_acc = []
        for icol in range(ncols - 1):
            if icol in [0, 1]:
                if weights[0] < 1e-3:
                    continue
            elif weights[icol - 1] < 1e-3:
                continue
            cols_taken_into_acc.append(icol)
        cols_taken_into_acc = np.array(cols_taken_into_acc)

        df_subset = df[indices_for_label, :][:, cols_taken_into_acc]
        means, stds = _truncated_means_and_stds(df_subset, truncate)
        tol = max_n_stdev_around_mean * stds[:]
        for irow in indices_for_label:
            for icol in cols_taken_into_acc:
                if np.abs(df[irow, icol] - means[icol]) > tol[icol]:
                    df[irow, -1] = -(df[irow, -1] + rm_obs_label_offset)
                    break

    return df


def _get_new_labels_for_deviating_obs(
    df,
    max_num_iter=100,
    max_n_stdev_around_mean=2.0,
    trunc_perc=0.0,
    weights=None,
    verbose=False,
):
    """
    Detects and assign new labels to obs where abs(obs[i]-obs_mean)>tolerance

    Helper function for the filter_outliers_iterative routine
    """
    nrows, ncols = df.shape
    if weights is None:
        weights = np.ones(ncols - 2)
    # Set any negative weight value to zero
    weights = np.where(weights < 0, 0.0, weights)

    unique_labels = np.unique(df[:, -1])
    rm_obs_label_offset = get_rm_obs_label_offset(unique_labels)

    n_removed_tot = 0
    for i in range(max_num_iter):
        if verbose:
            print("    > Refining scan, iteration ", i + 1, ": ", end=" ")
            start_time = time.time()
        n_removed_old = np.count_nonzero(df[:, -1] < -1)
        df = _get_new_labels_for_deviating_obs_one_iter(
            df,
            max_n_stdev_around_mean,
            rm_obs_label_offset,
            truncate=trunc_perc,
            weights=weights,
        )
        n_removed_new = np.count_nonzero(df[:, -1] < -1)
        n_removed_this_iter = n_removed_new - n_removed_old
        n_removed_tot = n_removed_tot + n_removed_this_iter
        if verbose:
            print("* rm ", n_removed_this_iter, end=" ")
            print(" stations, ", n_removed_tot, " so far. ", end=" ")
            print("Took ", time.time() - start_time, "s")
        if n_removed_this_iter == 0:
            break
    return df[:, -1]


def filter_outliers_iterative(
    df,
    skip,
    weights,
    trunc_perc=0.25,
    max_num_refine_iter=1000,
    max_n_stdev_around_mean=2,
    **kwargs
):
    df = df.rename(columns={"cluster_label": "original_cluster_label"})
    df["cluster_label"] = _get_new_labels_for_deviating_obs(
        df.drop(skip, axis=1).to_numpy(),
        max_num_iter=max_num_refine_iter,
        max_n_stdev_around_mean=max_n_stdev_around_mean,
        trunc_perc=trunc_perc,
        weights=weights,
        verbose=logger.getEffectiveLevel() == logging.DEBUG,
    ).astype(int)
    return df


# GLOSH outlier removal method
def filter_outliers_glosh(df, db, **kwargs):
    df["original_cluster_label"] = df["cluster_label"]
    df["GLOSH"] = db.outlier_scores_
    threshold = min(0.25, df[df["cluster_label"] > -1]["GLOSH"].quantile(0.75))
    outliers_index = (df["cluster_label"] > -1) & (df["GLOSH"] > threshold)
    rm_obs_label_offset = get_rm_obs_label_offset(
        df["cluster_label"].to_numpy()
    )
    df.at[outliers_index, "cluster_label"] = -(
        df["cluster_label"][outliers_index] + rm_obs_label_offset
    )
    return df


# Routines related to the LOF outlier removal method
def get_local_outlier_factors(df, distance_matrix, calc_per_cluster=False):
    # See <https://scikit-learn.org/stable/modules/generated/
    # sklearn.neighbors.LocalOutlierFactor.html#>
    unique_labels = df["cluster_label"].unique()
    all_lof_values = np.empty(len(df.index))
    all_lof_values[:] = np.nan
    # Is it better to do this in a per-cluster bases or
    # with the dataset as a whole? Maybe, with such a small
    # n_neighbors, the results won't change much.
    if calc_per_cluster:
        for label in unique_labels:
            if label < 0:
                continue
            indices_mask = df["cluster_label"] == label
            indices = df.index[indices_mask]
            reduced_dist_matrix = distance_matrix[indices, :][:, indices]
            clf = LocalOutlierFactor(
                n_neighbors=min(3, len(indices)), metric="precomputed"
            )
            clf.fit_predict(reduced_dist_matrix)
            all_lof_values[indices] = clf.negative_outlier_factor_
    else:
        indices_mask = df["cluster_label"] > -1
        indices = df.index[indices_mask]
        reduced_dist_matrix = distance_matrix[indices, :][:, indices]
        clf = LocalOutlierFactor(n_neighbors=3, metric="precomputed")
        clf.fit_predict(reduced_dist_matrix)
        all_lof_values[indices] = clf.negative_outlier_factor_
    return all_lof_values


def filter_outliers_lof(df, distance_matrix, **kwargs):
    df["original_cluster_label"] = df["cluster_label"]
    df["LOF"] = get_local_outlier_factors(df, distance_matrix)
    outliers_index = (df["cluster_label"] > -1) & (df["LOF"] < -1.5)
    rm_obs_label_offset = get_rm_obs_label_offset(
        df["cluster_label"].to_numpy()
    )
    df.at[outliers_index, "cluster_label"] = -(
        df["cluster_label"][outliers_index] + rm_obs_label_offset
    )
    return df


# "reclustering" outlier removal method
@suspendlogging
def filter_outliers_reclustering(df, distance_matrix, **kwargs):
    n_iter = 0
    n_removed = 0
    n_noise = np.count_nonzero(df["cluster_label"] < 0)
    while n_noise != 0:
        noise_mask = df["cluster_label"] < 0
        df_noise = df[noise_mask]

        i_valid_obs = np.nonzero((~noise_mask).values)[0]
        reduced_dist_matrix = distance_matrix[i_valid_obs, :][:, i_valid_obs]

        df = df.drop(df[noise_mask].index)
        df = df.drop(["cluster_label"], axis=1)
        df_rec = run_clustering_on_df(
            df,
            outlier_rm_method=None,
            sort_by_cluster_size=False,
            distance_matrix=reduced_dist_matrix,
            **kwargs,
        )
        n_noise = np.count_nonzero(df_rec["cluster_label"] < 0)
        df = pd.concat([df_noise, df_rec]).sort_index()
        n_iter += 1
        n_removed += n_noise
        logger.debug("        * Iter #%d: %d new noise pts", n_iter, n_noise)
    logger.debug("    > Done with reclustering. %d removed obs.", n_removed)
    return df


# General outlier removal routine, which calls the specific ones defined above
def filter_outliers(df, db, outlier_rm_method, distance_matrix, **kwargs):
    allowed_methods = dict(
        dbscan=["lof", "iterative", "reclustering"],
        hdbscan=["glosh", "lof", "iterative", "reclustering"],
    )
    clustering_method = "dbscan"
    if hasattr(db, "outlier_scores_"):
        clustering_method = "hdbscan"
    outlier_rm_method = outlier_rm_method.lower()
    if outlier_rm_method not in allowed_methods[clustering_method]:
        raise ValueError(
            'ERROR: Outlier removal method "{}" '.format(outlier_rm_method)
            + " not supported for {}.".format(clustering_method)
        )
    logger.debug('    > Running outlier rmv method "%s"', outlier_rm_method)
    tstart = time.time()
    if outlier_rm_method == "glosh":
        rtn = filter_outliers_glosh(df, db=db)
    elif outlier_rm_method == "lof":
        rtn = filter_outliers_lof(df, distance_matrix=distance_matrix)
    elif outlier_rm_method == "reclustering":
        rtn = filter_outliers_reclustering(df, distance_matrix, **kwargs)
    else:
        rtn = filter_outliers_iterative(df, **kwargs)
    logger.debug(
        "      * Done with outlier removal. Elapsed: %.1fs",
        time.time() - tstart,
    )
    return rtn


def sort_df_by_cluster_size(df):
    # Sort df so that clusters with more members are put at the top of the
    # dataframe. The exception if the "-1" label, whichm if present, will
    # always remain at the top. Handy if results are to be plotted.

    # The labels may have been reassigned if the df was passed through
    # an outlier removal routine. Let's keep track of the original ones.
    # This will help keeping obs removed from a cluster by the outlier
    # removed routine close to their origin cluster.
    original_cluster_label_col = "cluster_label"
    if "original_cluster_label" in df.columns:
        original_cluster_label_col = "original_cluster_label"

    # Temporarily add a column with cluster sizes for sorting
    unique_labels, member_counts = np.unique(
        df[original_cluster_label_col], return_counts=True
    )
    label2count_dict = {
        label: count for label, count in zip(unique_labels, member_counts)
    }
    df["parent_cluster_size"] = [
        label2count_dict[label] for label in df[original_cluster_label_col]
    ]
    # Sort dataframe according to frequency, but keep -1 at the top
    df = pd.concat(
        [
            df[df[original_cluster_label_col] == -1],
            df[df[original_cluster_label_col] != -1].sort_values(
                [
                    "parent_cluster_size",
                    original_cluster_label_col,
                    "cluster_label",
                ],
                ascending=[False, True, True],
            ),
        ]
    )

    # Reassign labels so that fewer members leads to larger labels.
    # Mind that the df unique method does not sort, so this method won't
    # mess up the sorting performed above.
    unique_labels = df[original_cluster_label_col].unique()
    labels_old2new = dict(
        (old, new)
        for new, old in enumerate(l for l in unique_labels if l >= 0)
    )
    labels_old2new[-1] = -1

    if original_cluster_label_col == "original_cluster_label":
        # This happens if some of the outlier removal routines are run.
        # If an observation is removed by such methods, it is helpful to
        # be able to know which cluster the obs was removed from.
        # Our convention throughout the code is to re-label the removed
        # obs according to:
        #
        #            new_label = -(old_label + rm_obs_label_offset),
        #
        # with the positive integer "rm_obs_label_offset" calculated
        # as in the get_rm_obs_label_offset function.
        #
        # Such a convention allows us to keep track of cluster origin for
        # removed obs without adding a new column to the dataframe. One
        # could think of just multiplying original labels by -1, but that
        # approach has two main problems:
        #    * Obs removed from label 0 would not be flagged
        #    * Obs removed from label 1 would be marked -1, but -1 is
        #      reserved for obs removed by the clustering itself
        rm_obs_label_offset = get_rm_obs_label_offset(
            df["cluster_label"].to_numpy()
        )

        # When building the labels_old2new dictionary after having reordered
        # the entries in the dataframe, we need to account for and keep using
        # this convention.
        for label in df["cluster_label"][df["cluster_label"] < -1].unique():
            cluster_it_was_rm_from = -np.rint(
                label + rm_obs_label_offset
            ).astype(int)
            labels_old2new[label] = -(
                labels_old2new[cluster_it_was_rm_from] + rm_obs_label_offset
            )

    new_labels = np.array([labels_old2new[old] for old in df["cluster_label"]])
    df["cluster_label"] = new_labels

    return df.drop("parent_cluster_size", axis=1).reset_index(drop=True)


def run_clustering_on_df(
    df,
    method="hdbscan",
    distance_matrix=None,
    skip=["id", "time_utc"],
    weights_dict={},
    eps=15,  # eps applies only to dbscan
    min_cluster_size=3,  # min_cluster_size applies only to hdbscan
    min_samples=3,
    n_jobs=-1,
    outlier_rm_method=None,
    max_num_refine_iter=50,
    max_n_stdev_around_mean=2.0,
    trunc_perc=0.25,
    sort_by_cluster_size=True,
    **kwargs
):

    method = method.lower()
    # Compute clustering using DBSCAN or HDBSCAN
    if method not in ["dbscan", "hdbscan", "rsl", "optics"]:
        raise NotImplementedError('Method "{}" not available.'.format(method))
    if len(df.index) == 0:
        raise ValueError("Dataframe has no rows")

    # Set weights to be used in the metrics for the various
    # generalised distances. The distances used in the metrics
    # will be used_dist(i) = weight(i)*real_dist(i)
    weights = get_weights_array(df, weights_dict)

    # We will not do any df = StandardScaler().fit_transform(df),
    # as we'll use a metric based on earth distances

    if distance_matrix is None:
        # My tests indicate that passing a pre-computed distance matrix to
        # dbscan can be up to 2.5x faster than passing a metrics function (if
        # they both are written in pure python) to fit df. If they are both
        # written in fortran and interfaced via f2py3, or then written in
        # python but jit-compiled with numba, then the relative speed up
        # can reach up to 120x.
        num_threads = -1
        if "num_threads" in kwargs:
            num_threads = kwargs["num_threads"]
        distance_matrix = calc_distance_matrix(
            # Drop columns that won't be used in the clustering
            df.drop(skip, axis=1),
            weights,
            num_threads,
        )

    # Running clustering with the computed distance matrix
    logger.debug('    > Running clustering method "{}"...'.format(method))
    tstart = time.time()
    if method == "dbscan":
        db = DBSCAN(
            eps=eps,
            min_samples=min_samples,
            metric="precomputed",
            n_jobs=n_jobs,
        ).fit(distance_matrix)
        core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
        core_samples_mask[db.core_sample_indices_] = True
    elif method == "hdbscan":
        # For more info on the parameters, see
        # <https://hdbscan.readthedocs.io/en/latest/parameter_selection.html>
        db = HDBSCAN(
            min_samples=min_samples,
            min_cluster_size=min_cluster_size,
            metric="precomputed",
            core_dist_n_jobs=n_jobs,
            allow_single_cluster=True,
            # Default cluster_selection_method: 'eom'. Sometimes it leads to
            # clusters that are too big. Using 'leaf' seems better.
            cluster_selection_method="leaf",
            # cluster_selection_epsilon=7.5,
        ).fit(distance_matrix)
        # core_samples info is not available with hdbscan
        core_samples_mask = None
    elif method == "rsl":
        db = RobustSingleLinkage(
            # cut: The reachability distance value to cut the cluster
            #      heirarchy at to derive a flat cluster labelling.
            cut=eps,  # default=0.4
            # Reachability distances will be computed with regard to the
            # k nearest neighbors.
            k=min_samples,  # default=5
            # Ignore any clusters in the flat clustering with size less
            # than gamma, and declare points in such clusters as noise points.
            gamma=min_cluster_size,  # default=5
            metric="precomputed",
        ).fit(distance_matrix)
        # core_samples info is not available with rsl
        core_samples_mask = None
    elif method == "optics":
        db = OPTICS(
            min_samples=min_samples,
            min_cluster_size=min_cluster_size,
            n_jobs=n_jobs,
            metric="precomputed",
        ).fit(distance_matrix)
        # core_samples info is not available with optics
        core_samples_mask = None
    logger.debug(
        "      * Done with {0}. Elapsed: {1:.1f}s".format(
            method, time.time() - tstart
        )
    )
    # Update df with cluster label info. It is important that this is done
    # right before calling _get_new_labels_for_deviating_obs, as the
    # _get_new_labels_for_deviating_obs functions expects 'cluster_label'
    # to be the last column in the dataframe
    df["cluster_label"] = db.labels_

    # Refine clustering if requested
    # It is important to have 'cluster_label' as the last column
    # when running the iterative refine routine
    if outlier_rm_method:
        df = filter_outliers(
            df,
            db=db,
            outlier_rm_method=outlier_rm_method,
            # Args that apply only to LOF
            distance_matrix=distance_matrix,
            # Args that only apply for the iterative method
            skip=skip,
            max_num_refine_iter=max_num_refine_iter,
            max_n_stdev_around_mean=max_n_stdev_around_mean,
            trunc_perc=trunc_perc,
            weights=weights,
            method=method,
            weights_dict=weights_dict,
            eps=eps,
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            n_jobs=n_jobs,
        )

    # core_samples_mask is only returned by dbscan, not by hdbscan
    if core_samples_mask is not None:
        df["is_core_point"] = core_samples_mask

    # Calculate silhouette scores and update df with this.
    # We'll be reordering the dataframe later, and calculating the score
    # before doing that avoids the need to also reorder the dist matrix.
    df["silhouette_score"] = get_silhouette_samples(df, distance_matrix)

    if sort_by_cluster_size:
        df = sort_df_by_cluster_size(df)

    # Remove aux columns that are no longe rneeded
    try:
        df = df.drop("original_cluster_label", axis=1)
    except (KeyError):
        pass

    return df


def cluster_netatmo_obs(
    config,
    dtg=None,
    sort_by_cluster_size=False,
    return_list_of_removed_stations=False,
    **kwargs
):
    if dtg is None:
        dtg = config.dtgs[0]
    logger.debug("Reading data for DTG=%s...", dtg)
    start_read_data = time.time()
    if return_list_of_removed_stations:
        df, rmvd_stations = read_netatmo_data_for_dtg(
            dtg,
            rootdir=config.data_rootdir,
            return_list_of_removed_stations=True,
        )
    else:
        df = read_netatmo_data_for_dtg(dtg, rootdir=config.data_rootdir)
    end_read_data = time.time()
    logger.debug(
        "Done reading data for DTG={}. Elapsed: {:.1f}s".format(
            dtg, end_read_data - start_read_data
        )
    )

    time_start_clustering = time.time()
    logger.debug("Performing clustering...")
    df = run_clustering_on_df(
        df=df,
        method=config.clustering_method,
        weights_dict=config.get("obs_weights", {}),
        eps=config.get("eps", 10),
        min_cluster_size=config.get("min_cluster_size", 5),
        min_samples=config.get("min_samples", 5),
        outlier_rm_method=config.outlier_removal.method,
        max_num_refine_iter=config.get("outlier_removal.max_n_iter", 1000),
        max_n_stdev_around_mean=config.get("outlier_removal.max_n_stdev", 2),
        # We only need this sorting if we're plotting
        sort_by_cluster_size=sort_by_cluster_size,
        **kwargs,
    )
    time_end_clustering = time.time()
    logger.debug(
        "Done with clustering. Elapsed: {}s".format(
            np.round(time_end_clustering - time_start_clustering, 2)
        )
    )

    if return_list_of_removed_stations:
        return df, rmvd_stations
    else:
        return df


def report_clustering_results(df):
    n_obs = len(df.index)
    noise_data_df = df[df["cluster_label"] < 0]
    n_noise_clusters = len(noise_data_df["cluster_label"].unique())
    noise_count = len(noise_data_df)
    n_clusters = len(df["cluster_label"].unique()) - n_noise_clusters
    n_accepted = n_obs - noise_count
    silhouette_score = df["silhouette_score"].mean(skipna=True)

    logger.info("Total number of parsed observations: %d", n_obs)
    logger.info("Estimated number of clusters: %d", n_clusters)
    logger.info("Estimated number of accepted obs: %d", n_accepted)
    logger.info(
        "Estimated number of noise obs: %d (%.1f%% rejection rate)",
        noise_count,
        100.0 * noise_count / n_obs,
    )
    logger.info("Mean silhouette score: {:.3f}".format(silhouette_score))



#!/usr/bin/env python3
import numpy as np
import time
import numba
from numba import njit, prange
import logging
import atexit

logger = logging.getLogger(__name__)


@njit(cache=True)
def _matrix_index_to_data_index(order, i, j):
    """Convert (i, j) index into single data index for a sparse dist matrix

    Given a square matrix M of order "order" which also satisfies
                M(i, j) = M(j, i) and M(i, i) = 0
    and a linear array, say "M_data" that stores only the independent elements
    of M, return the index i_data such that
                M(i, j) = M_data(i_data)

    This routine assumes
                        0 <= i,j < order
    To make using numba.jit easier, no checks
    are performed on indices, but, if there's a possibility that users will
    have access to this functionality, then one should ideally expose a
    wrapper that performs such checks before calling this function.

    This function is used in the SparseSymmZeroDiagMatrix class, where
    consistency between the arguments is enforced.
    """

    # Start by enforcing M(i, j) = M(j, i)
    if i > j:
        i, j = j, i

    # n_indep_els_before_row: Total number of independent elements
    # located in the rows prior to row=i for the matrix M
    n_indep_els_before_row = i * order - (i * (i + 1)) // 2

    return n_indep_els_before_row + (j - i - 1)


@njit(cache=True)
def _data_index_to_matrix_index(n, i_data):
    """Convert linear data index into matrix indices

    Given a square matrix M of order "n" which also satisfies
                M(i, j) = M(j, i) and M(i, i) = 0
    and a linear array, say "M_data" that stores only the independent elements
    of M, return the matrix indices (i, j) such that
                M(i, j) = M_data(i_data)
    This is the inverse of what is done in _matrix_index_to_data_index.

    This routine assumes
                        0 <= i_data < n(n-1)/2
    To make using numba.jit easier, no checks are performed on i_data,
    but, if there's a possibility that users will have access to this
    functionality, then one should ideally expose a wrapper that performs
    such checks before calling this function.

    This function is used in the SparseSymmZeroDiagMatrix class, where
    consistency between the arguments is enforced.
    """
    i = (
        int(0.5 * ((2 * n + 1) - np.sqrt((2 * n + 1) ** 2 - 8 * (n + i_data))))
        - 1
    )
    j = (1 + i_data + i) - (i * (2 * n - i - 1)) // 2
    return i, j


@njit(cache=True)
def _to_dense(data, n, dtype=None, copy=None, order=None):
    """Dense form of sparse matrix M with data 'data', M(i,j)=M(j,i), M(i,i)=0

    Takes an array 'data' and an integer 'n' and constructs a full (dense)
    representation of a square matrix M of order 'n' such that
                       M(i,j)=M(j,i), M(i,i)=0
    and whose independent elements are stored sequentially in 'data'.

    Assumes that len(data) = n*(n - 1)/2.
    No checks, however, are made on len(data). If you want to provide this
    functionality in an external API, please expose a wrapper to this function
    where this is asserted.

    This function is used in the SparseSymmZeroDiagMatrix class, where
    consistency between the arguments is enforced.

    The args dtype, copy and order are ignored, but this interface was
    necessary in order to be able to use the optics clustering method.
    """
    dense = np.zeros(shape=(n, n))
    for i in range(n - 1):
        for j in range(i + 1, n):
            data_val = data[_matrix_index_to_data_index(n, i, j)]
            dense[i, j] = data_val
            dense[j, i] = data_val
    return dense


class SparseSymmZeroDiagMatrix:
    """Sparse, square, symmetric matrix with zeros along the diagonal

    The order N of the (N, N) matrix is calculated from the passed data array.
    The diagonal elements are not stored (assumed to be all zero), and only
    half of the remaining elements are stored, as M(i, j) = M(j, i).
    """

    def __init__(self, data):
        # N: Order N of the NxN square matrix M(i, j)
        # n_indep: Number of independent elements
        #     Since M is symmetric and diagnonal elements are always zero,
        #     n_indep = (N**2 - N)/2
        n_indep = len(data)
        N = 0.5 * (1.0 + np.sqrt(1.0 + 8.0 * n_indep))
        nearest_int_N = int(np.rint(N))
        if abs(N - nearest_int_N) > 1e-8:
            nearest_n_indep = int(0.5 * nearest_int_N * (nearest_int_N - 1))
            msg = "len(data)={0} incompatible with an NxN distance matrix: "
            msg += "{1} elements needed for a {2}x{2} matrix (closest option)"
            msg = msg.format(n_indep, nearest_n_indep, nearest_int_N)
            raise ValueError(msg)
        self._order = int(np.rint(N))
        self._data = np.array(data)

    @property
    def order(self):
        return self._order

    @property
    def shape(self):
        return (self.order, self.order)

    @property
    def data(self):
        return self._data

    def matrix_index_to_data_index(self, i, j):
        """Returns an integer I such that self.data[I] == self[i, j]"""
        if i >= self.order:
            raise IndexError(
                "index {} is out of bounds for axis 1 with size {}".format(
                    i, self.order
                )
            )
        if j >= self.order:
            raise IndexError(
                "index {} is out of bounds for axis 1 with size {}".format(
                    j, self.order
                )
            )
        return _matrix_index_to_data_index(self.order, i, j)

    def data_index_to_matrix_index(self, data_i):
        """Returns (i, j) such that self.data[data_i] == self[i, j]"""
        if data_i >= len(self.data):
            raise IndexError(
                "index {} is out of bounds for axis 1 with size {}".format(
                    data_i, self.data
                )
            )
        return _data_index_to_matrix_index(self.order, data_i)

    def __array__(self, dtype=None, copy=False, order=None):
        # See <https://numpy.org/devdocs/user/basics.dispatch.html>
        return _to_dense(self.data, self.order, dtype, copy, order)

    def __getitem__(self, ij_tuple):
        if any(isinstance(index, slice) for index in ij_tuple):
            # TODO: Implement a proper getitem for slices.
            #       This is fast but not memory-efficient,
            #       as it creates a full version of the array
            return self.__array__()[ij_tuple]
        else:
            if ij_tuple[0] == ij_tuple[1]:
                return 0.0
            else:
                return self.data[self.matrix_index_to_data_index(*ij_tuple)]

    def copy(self):
        # TODO: Make a proper copy implementation. This is a quick fix
        # because I need this for hdbscan.
        return np.asarray(self)


# Routines related to the calculation of the distance matrix
@njit(cache=True)
def haversine_distance(point1, point2):
    lat1 = np.deg2rad(point1[0])
    lat2 = np.deg2rad(point2[0])
    dlat = lat2 - lat1
    dlon = np.deg2rad(point2[1] - point1[1])
    a = (
        np.sin(dlat * 0.5) ** 2
        + np.cos(lat1) * np.cos(lat2) * np.sin(dlon * 0.5) ** 2
    )
    # earth_mean_diameter = 12742.0176 Km
    rtn = 12742.0176 * np.arcsin(np.sqrt(a))
    return rtn


@njit(cache=True)
def get_obs_norm_factors(obs_values):
    stdev = np.std(obs_values)
    obs_mean = np.mean(obs_values)
    if (
        (abs(obs_mean) > 1e-5 and abs(stdev / obs_mean) < 1e-5)
        or abs(obs_mean) <= 1e-5
        and stdev < 1e-7
    ):
        rtn = 0.0
    else:
        rtn = 1.0 / stdev
    return rtn


@njit("f4[:](f8[:, :], f8[:])", parallel=True, cache=True)
def calc_distance_matrix_numba(df, weights):
    nrows, ncols = df.shape

    # Get normalisation factors so that observations in different
    # scales have comparable diff values
    # Remember: The first entry in weights refers to geo_dist and
    #           the first two columns of df are (lat, lon)
    obs_norm_factors = np.ones(ncols - 1, dtype=np.float32)
    for i in range(1, len(obs_norm_factors)):
        obs_norm_factors[i] = get_obs_norm_factors(df[:, i + 1])

    # weights can be used by the users to give extra or
    # lower weight to a given observation type. The weights
    # are applied to the normalised values of the obd diffs
    #
    # We are not normalising geodists. Therefore, the normalised obs
    # dists should have the same units as the geodists (km).
    # Let's assign 10 km/normalized_obs_unit_diff by default
    # This unit conversion is highly arbitrary...

    weights_internal = weights
    # Set any negative weight value to zero
    weights_internal = np.where(weights_internal < 0, 0.0, weights_internal)
    weights_internal *= obs_norm_factors

    n_dists = (nrows * (nrows - 1)) // 2
    rtn = np.zeros(n_dists, dtype=np.float32)
    # For better parallel performance, loop over single idist instead of doing
    # a nested "for i in prange(nrows): for j in range(i+1, nrows):".
    # If one wishes to reverse this, then use func _matrix_index_to_data_index
    # or note that idist = i*nrows - i*(i+1)//2 + (j - i - 1)
    # NB.: Initialising i and j as integers, otherwise numba v0.50.1 fails
    # silently and gives wrong results in python 3.8 with "parallel=True".
    # Strangely, the results are correct without initialisation when using
    # python 3.6 or when setting parallel=False.
    i, j = np.zeros(2, dtype=np.int64)
    for idist in prange(n_dists):
        i, j = _data_index_to_matrix_index(nrows, idist)
        rtn[idist] = weights_internal[0] * haversine_distance(
            df[i], df[j]
        ) + np.sum(
            np.sqrt((weights_internal[1:] * (df[j, 2:] - df[i, 2:])) ** 2)
        )

    return rtn


def calc_distance_matrix(df, weights, num_threads=-1):
    logger.debug("    > Calculating distance matrix...")
    tstart = time.time()

    if num_threads > 0:
        original_nthreads = numba.get_num_threads()
        numba.set_num_threads(num_threads)
        atexit.register(numba.set_num_threads, original_nthreads)

    rtn = SparseSymmZeroDiagMatrix(
        data=calc_distance_matrix_numba(df.to_numpy(), weights=weights)
    )

    logger.debug(
        "      * Done calculating distance matrix. "
        + "Elapsed: {:.1f}s".format(time.time() - tstart)
    )
    return rtn



import toml
from datetime import datetime, date
import os
import sys
import numpy as np
import logging
import pandas as pd
import functools
from dotmap import DotMap
import unittest
from pathlib import Path
from .dtgs import DtgContainer

logger = logging.getLogger(__name__)


class UndefinedValueType:
    """
    This class is defined such that any instance foo satisfies:
        (a) bool(foo) is False
        (b) If return_self_on_attr_error=True, then, for any attribute bar,
            foo.bar is foo. This implies that, for any N>1,
            foo.bar1.bar2.(...).barN is foo

    We will use instances of this class to mark things such as missing
    config file values, unspecified defaults in functions, etc. We don't
    want to use None in these cases, as None may be a valid arg value.
    """

    def __init__(
        self,
        name="<UndefinedValueType object>",
        return_self_on_attr_error=False,
    ):
        self._name = name
        self._return_self_on_attr_error = return_self_on_attr_error

    # Define __setstate__ and __getstate__ to handle serialisation (pickling)
    # and avoid recursion errors (UndefinedConfigValue is a recursive object)
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    def __getattr__(self, item):
        if self._return_self_on_attr_error:
            return self
        else:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, item
                )
            )

    def __bool__(self):
        return False

    def __repr__(self):
        return self._name

    __str__ = __repr__


UndefinedConfigValue = UndefinedValueType(
    name="UndefinedConfigValue", return_self_on_attr_error=True
)
NoDefaultProvided = UndefinedValueType(name="NoDefaultProvided")


class UndefinedConfigValueError(Exception):
    pass


class ConfigParsingError(Exception):
    pass


class _ConfigDict(DotMap):
    """
    Acts like a normal dictionary, except for:
        (a) Accepts getting/setting values using dot notation (inherited from
            the dotmap package). This means that instance.key is equivalent
            to instance[key].
        (b) Returns UndefinedConfigValue if a key is not present instead
            of raising KeyError.
    """

    def __init__(self, *args, **kwargs):
        kwargs["_dynamic"] = False
        super(_ConfigDict, self).__init__(*args, **kwargs)

    # Handling value if retrieving as dict key
    def __getitem__(self, item):
        return self.get(item, default=UndefinedConfigValue)

    # Handling value if using dot notation
    # __getattr__ is only called if the attribute does NOT exist
    def __getattr__(self, item):
        return self.get(item, default=UndefinedConfigValue)


def _rgetattr(obj, attr, *args):
    """
    Works similarly to getattr, but:
        (a) Handles dotted attr strings. Adapted from:
            <https://stackoverflow.com/questions/31174295/
             getattr-and-setattr-on-nested-subobjects-chained-properties>
        (b) Converts input/output dictionaries into objects of our module's
            _ConfigDict class
    """

    def _getattr(obj, attr):
        if type(obj) == dict or isinstance(obj, DotMap):
            obj = _ConfigDict(obj)
        rtn = getattr(obj, attr, *args)
        if isinstance(rtn, DotMap):
            rtn = _ConfigDict(rtn)
        return rtn

    return functools.reduce(_getattr, [obj] + attr.split("."))


def _parse_dtg_entires(dtgs_config):
    try:
        cycle_length = dtgs_config["cycle_length"]
    except (KeyError):
        cycle_length = "3H"
        logger.warning("Missing cycle_length in config. Setting it to 3H.")

    dtg_range_opt_labels = set(["start", "end"])
    if "list" in dtgs_config.keys():
        if not dtg_range_opt_labels.isdisjoint(dtgs_config.keys()):
            logger.warning(
                'Found DTG options "list" and also at least '
                + 'one of [%s]. Parsing only values passed in "list".',
                ", ".join(dtg_range_opt_labels),
            )
        try:
            dtgs = DtgContainer(dtgs_config["list"], cycle_length=cycle_length)
        except (ValueError) as e:
            logger.exception(e)
            raise ValueError("Bad 'dtg.list' or 'dtg.cycle_length' config.")
    else:
        try:
            start = dtgs_config["start"]
            end = dtgs_config["end"]
        except (KeyError) as e:
            logger.exception(e)
            msg = " Either 'dtgs.list' or both 'dtgs.start' and "
            msg += "'dtgs.end' must specified in config file."
            raise UndefinedConfigValueError(msg)
        try:
            dtgs = DtgContainer(
                start=dtgs_config["start"],
                end=dtgs_config["end"],
                cycle_length=cycle_length,
            )
        except (ValueError, TypeError) as e:
            logger.exception(e)
            raise ValueError("Bad 'dtg.start/end/cycle_length' config.")

    return dtgs


def _validate_value(
    optname,
    optval=None,
    default=None,
    choices=None,
    maxval=None,
    minval=None,
    astype=None,
):
    if optval is None:
        optval = default
    # Pre-parsing choices. They may sometimes be functions.
    try:
        choices = choices()
    except (TypeError):
        pass
    # Cast to desired type, if passed
    if astype is not None:
        try:
            optval = astype(optval)
            if default is not None:
                default = astype(default)
            if maxval is not None:
                maxval = astype(maxval)
            if minval is not None:
                minval = astype(minval)
            if choices is not None:
                choices = list(map(astype, choices))
        except (TypeError, ValueError) as e:
            logger.exception(e)
            msg = 'Could not parse "{0}" using "astype={1}"'.format(
                optname, astype
            )
            raise ValueError(msg)

    if (choices is not None) and (optval not in choices):
        raise ValueError(
            '{0} "{1}" not valid. Valid choices are: {2}'.format(
                optname, optval, ", ".join(map(str, choices))
            )
        )
    if (minval is not None) and optval < minval:
        raise ValueError(
            "{0} value ({1}) smaller than the allowed min ({2})".format(
                optname, optval, minval
            )
        )
    if (maxval is not None) and optval > maxval:
        raise ValueError(
            "{0} value ({1}) larger than the allowed max ({2})".format(
                optname, optval, maxval
            )
        )
    return optval


def _validate_outlier_rm_method_config(clustering_method, config):
    if clustering_method in ["dbscan", "rsl"]:
        valid_out_rm_methods = ["lof", "iterative", "reclustering"]
    elif clustering_method == "hdbscan":
        valid_out_rm_methods = ["glosh", "lof", "iterative", "reclustering"]
    else:
        logger.warning(
            "Cannot verify validity of outlier_removal for  "
            + 'clustering method "%s". Passing on "as is"',
            clustering_method,
        )
        return config

    if config["method"] not in valid_out_rm_methods:
        raise ValueError(
            (
                'outlier_removal method "{}" not valid for '
                + 'clustering_method "{}". Valid values are: {}'
            ).format(
                config["method"],
                clustering_method,
                ", ".join(valid_out_rm_methods),
            )
        )

    if config["method"] == "iterative":
        recognised_opts = dict(
            max_n_iter=dict(default=100, minval=0),
            max_n_stdev=dict(default=2, minval=0),
        )
        for optname in config:
            if optname == "method":
                continue
            try:
                optconfigs = recognised_opts[optname]
            except (KeyError):
                logger.warning(
                    'Config opt "%s" (in %s->outlier_removal->iterative)'
                    + ' not recognised. Passing on "as is"',
                    optname,
                    clustering_method,
                )
                continue
            try:
                optval = config[optname]
            except (KeyError):
                optval = None
            config[optname] = _validate_value(
                optname="{} ({} -> outlier_removal -> iterative)".format(
                    optname, clustering_method
                ),
                optval=optval,
                **optconfigs
            )

    return config


class ParsedConfig:
    def __init__(
        self,
        file=Path().resolve() / "config.toml",
        global_default=NoDefaultProvided,
    ):
        self._file = Path(file)
        # Keep the config's raw data, mostly for debugging purposes
        self._raw = toml.load(self._file.resolve())

        # Directory where NetAtmo data is supposed to be found
        try:
            rootdir = self._raw["general"]["data_rootdir"]
            # expandvars adds support to envvars as part of data_rootdir
            self.data_rootdir = Path(os.path.expandvars(rootdir))
        except (KeyError):
            raise UndefinedConfigValueError(
                "data_rootdir not specified in config file"
            )

        self.global_default = global_default

        # Define some metadata about recognised clustering methods
        # and respective options. This will help parsing.
        self._recog_ops_for_cluster_methods = {}

        self._recog_ops_for_cluster_methods["dbscan"] = dict(
            eps=dict(default=10, minval=0),
            min_samples=dict(default=5, minval=1),
            outlier_removal={},  # This will be parsed separately
            obs_weights=dict(default={}, astype=dict),
        )
        self._recog_ops_for_cluster_methods["hdbscan"] = dict(
            min_samples=dict(default=5, minval=1),
            min_cluster_size=dict(default=5, minval=1),
            max_iter_refine=dict(default=100, minval=0),
            stdev_refine=dict(default=2, minval=0),
            obs_weights=dict(default={}, astype=dict),
            eps=dict(default=10, minval=0),
            outlier_removal={},  # This will be parsed separately
        )
        self._recog_ops_for_cluster_methods["rsl"] = dict()

        # Now parse input data (validate choices, etc)
        #
        # Set self._clustering_method to the chosen clustering_method
        # and set self._parsed to contain the parsed configs for
        # self._clustering_method. self.get will return values from
        # self._parsed.
        try:
            self.set_clustering_method(
                self._raw["general"]["clustering_method"]
            )
        except (KeyError):
            raise UndefinedConfigValueError(
                'clustering_method not defined under "general" in config file'
            )
        # Make sure at least one DTG is defined and that
        # defined DTG(s) are valid
        try:
            dtgs_config_raw = self._raw["general"]["dtgs"]
        except (KeyError):
            raise UndefinedConfigValueError(
                "DTG(s) not specified in config file!"
            )
        self._dtgs = _parse_dtg_entires(dtgs_config_raw)

        # Validate options that are valid for each individual command
        self.commands = _ConfigDict(dict(cluster={}, select={}, apps={},))
        def_stat_rejec_tol = 0.15
        try:
            self.commands.select.station_rejection_tol = _validate_value(
                "commands --> select --> station_rejection_tol",
                self._raw["commands"]["select"]["station_rejection_tol"],
                default=def_stat_rejec_tol,
                minval=0,
                maxval=1,
                astype=type(1.0),
            )
        except (KeyError):
            self.commands.select.station_rejection_tol = def_stat_rejec_tol

    @property
    def clustering_method(self):
        return self._clustering_method

    @property
    def dtgs(self):
        return self._dtgs

    @property
    def recognised_clustering_methods(self):
        return self._recog_ops_for_cluster_methods.keys()

    def set_clustering_method(self, method):
        self._clustering_method = method
        self._parsed = self._parse_selected_clustering_method_opts()

    def _parse_selected_clustering_method_opts(self):
        try:
            raw = self._raw["clustering_method"][self.clustering_method]
        except (KeyError):
            if self.global_default is NoDefaultProvided:
                raise UndefinedConfigValueError(
                    'No "clustering_method.{}"'.format(self.clustering_method)
                    + " block found in the config file"
                )
            else:
                logger.warning(
                    """
                    No "clustering_method.%s" block found in the config file.
                    Using the global default (= %s) for the whole config.
                    """,
                    self.clustering_method,
                    self.global_default,
                )
            return self.global_default
        try:
            opts = self._recog_ops_for_cluster_methods[self.clustering_method]
        except (KeyError):
            logger.warning(
                """
                No sanity checks for clustering method "{0}" are available
                at the moment. The whole "clustering_method.{0}" config file
                block will be used "as is".
                """.format(
                    self.clustering_method
                )
            )
            return raw

        parsed = {}
        for attr, attr_value in raw.items():
            try:
                if attr == "outlier_removal":
                    parsed[attr] = _validate_outlier_rm_method_config(
                        self.clustering_method, attr_value
                    )
                else:
                    opts = self._recog_ops_for_cluster_methods[
                        self.clustering_method
                    ][attr]
                    parsed[attr] = _validate_value(attr, attr_value, **opts)
            except (KeyError):
                logger.warning(
                    'Config opt "%s" not recognised by parser for '
                    + 'clustering method "%s". '
                    + 'Passing on "as is".',
                    attr,
                    self.clustering_method,
                )
                parsed[attr] = attr_value
        return parsed

    # Define __setstate__ and __getstate__ to handle serialisation (pickling)
    # and avoid recursion errors
    def __setstate__(self, state):
        self.__dict__ = state

    def __getstate__(self):
        return self.__dict__

    # Handling the getter method
    # __getattr__ is only called if the attribute does NOT exist
    def __getattr__(self, item):
        return self.get(item)

    def get(self, item, default=NoDefaultProvided):
        """Return parsed config values for the active clustering method"""
        # Re-format item so that the retrievals
        #            instance.get(['at1', 'at2', (...), 'atn']),
        #            instance.get(['at1.at2.(...).atn']),
        #            instance.at1.at2.(...).atn
        # are equivalent.
        if not isinstance(item, str):
            try:
                item = ".".join(iter(item))
            except (TypeError):
                msg = (
                    "The attribute(s) to be retrieved must be specified "
                    + "either as a string or as an iterable type. "
                    + 'Got "{}" instead (type={})'
                ).format(item, type(item))
                raise TypeError(msg)

        try:
            rtn = _rgetattr(self._parsed, item)
            if rtn is UndefinedConfigValue:
                # Handle this in the except block below
                raise AttributeError("{} is UndefinedConfigValue".format(item))
            else:
                return rtn
        except (AttributeError):
            if default is not NoDefaultProvided:
                logger.debug(
                    'No "%s" in config. Using passed default (= %s)',
                    item,
                    default,
                )
                return default
            elif self.global_default is not NoDefaultProvided:
                logger.warning(
                    'No "%s" in config for "%s". Using global_default (= %s)',
                    item,
                    self.clustering_method,
                    self.global_default,
                )
                return self.global_default
            else:
                raise UndefinedConfigValueError(
                    'Config has no value for "{}", '.format(item)
                    + "and no default has been passed."
                )


def read_config(config_path):
    logging.info("Reading config file %s", Path(config_path).resolve())
    config = ParsedConfig(
        file=config_path, global_default=UndefinedConfigValue
    )
    return config



import logging


class CustomFormatter(logging.Formatter):
    """
    Logging Formatter to add colors and count warning / errors

    Adapted from: <https://stackoverflow.com/questions/14844970/modifying-
        logging-message-format-based-on-message-logging-level-in-python3>
    """

    reset = "\u001b[0m"
    yellow = "\u001b[33m"
    red = "\u001b[31m"
    green = "\u001b[32m"

    FORMATS = {
        "DEFAULT": "%(asctime)s: %(message)s",
        logging.CRITICAL: (
            red
            + "%(asctime)s %(levelname)s(%(module)s: %(lineno)d): %(message)s"
            + reset
        ),
        logging.ERROR: red
        + "%(asctime)s %(levelname)s: "
        + reset
        + "%(message)s",
        logging.WARNING: yellow
        + "%(asctime)s %(levelname)s: "
        + reset
        + "%(message)s",
        logging.DEBUG: (
            green
            + "%(asctime)s %(levelname)s(%(module)s: %(lineno)d): "
            + reset
            + "%(message)s"
        ),
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno, self.FORMATS["DEFAULT"])
        formatter = logging.Formatter(log_fmt, datefmt="%Y-%m-%d %H:%M:%S")
        return formatter.format(record)


def get_logger(name, loglevel):
    logger = logging.getLogger(name)
    logger_handler = logging.StreamHandler()
    logger_handler.setFormatter(CustomFormatter())
    logging.basicConfig(
        level=logging.getLevelName(loglevel.upper()),
        handlers=[logger_handler],
    )
    return logger



try:
    # From python3.8
    from importlib.metadata import version
except (ModuleNotFoundError):
    # Prior to python3.8
    from importlib_metadata import version

try:
    __version__ = version(__name__)
except (PackageNotFoundError, ModuleNotFoundError):
    pass



#!/usr/bin/env python3
import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
import numpy as np
import datetime

# Plotly page on scatter plots on maps: https://plotly.com/python/scatter-plots-on-maps/
# See colorscales at https://plotly.com/python/builtin-colorscales/
# For animations, see https://plotly.com/python/animations/#animated-figures-with-graph-objects


def generate_single_frame(df, dataset_var, dt, frame_duration):

    # Controlling colour schemes
    if dataset_var in ["temperature"]:
        color_scale = px.colors.diverging.RdBu_r
    else:
        color_scale = px.colors.sequential.haline_r

    frame = {"data": None, "name": str(dt)}
    frame["data"] = go.Scattergeo(
        lon=df["lon"],
        lat=df["lat"],
        text=df[dataset_var],
        marker=dict(
            color=df[dataset_var],
            colorscale=color_scale,
            opacity=0.5,
            # size = 8,
            line=dict(color="black", width=0.25),
            colorbar=dict(
                titleside="right", ticks="outside", showticksuffix="last",
            ),
        ),
    )

    slider_step = {
        "args": [
            [str(dt)],
            {
                "frame": {"duration": frame_duration, "redraw": True},
                "mode": "immediate",
                "transition": {"duration": frame_duration},
            },
        ],
        "label": str(dt),
        "method": "animate",
    }

    return frame, slider_step


def init_fig_dict(dataset_var, frame_duration):
    # We will make the figure by constructing a dictionary and passing it fo the final plotly method
    fig_dict = {"data": [], "layout": {}, "frames": []}

    # Figure layout
    fig_dict["layout"]["title"] = "NetAtmo Data: {}".format(dataset_var)

    fig_dict["layout"]["width"] = 800
    fig_dict["layout"]["height"] = fig_dict["layout"]["width"]

    fig_dict["layout"]["geo"] = dict(
        scope="europe",
        resolution=110,
        projection_type="stereographic",
        showland=True,
        showcountries=True,
        lataxis=dict(
            showgrid=True,
            dtick=10,
            # range=(min_lat-1, max_lat+1)
        ),
        lonaxis=dict(
            showgrid=True,
            dtick=15,
            # range=(min_lon-1, max_lon+1)
        ),
    )

    fig_dict["layout"]["updatemenus"] = [
        {
            "buttons": [
                {
                    "args": [
                        None,
                        {
                            "frame": {
                                "duration": frame_duration,
                                "redraw": True,
                            },
                            "fromcurrent": True,
                            "transition": {
                                "duration": frame_duration,
                                "easing": "quadratic-in-out",
                            },
                        },
                    ],
                    "label": "Play",
                    "method": "animate",
                },
                {
                    "args": [
                        [None],
                        {
                            "frame": {"duration": 0, "redraw": True},
                            "mode": "immediate",
                            "transition": {"duration": 0},
                        },
                    ],
                    "label": "Pause",
                    "method": "animate",
                },
            ],
            "direction": "left",
            "pad": {"r": 10, "t": 87},
            "showactive": False,
            "type": "buttons",
            "x": 0.1,
            "xanchor": "right",
            "y": 0,
            "yanchor": "top",
        }
    ]

    sliders_dict = {
        "active": 0,
        "yanchor": "top",
        "xanchor": "left",
        "currentvalue": {
            "font": {"size": 20},
            "prefix": "DT: ",
            "visible": True,
            "xanchor": "right",
        },
        "transition": {"duration": frame_duration, "easing": "cubic-in-out"},
        "pad": {"b": 10, "t": 50},
        "len": 0.9,
        "x": 0.1,
        "y": 0,
        "steps": [],
    }

    return fig_dict, sliders_dict


# Make an animation
def make_map_animation(
    netatmo_data, dataset_var, min_dt_in_seconds=1.0, frame_duration=300
):
    # dts stand for dates and times
    dts = netatmo_data.keys()

    # Determine map boundaries and max/min plotted values
    min_lat = 360
    min_lon = 360
    max_lat = 0
    max_lon = 0
    minval_dataset_var = float("inf")
    maxval_dataset_var = -float("inf")
    for dt in dts:
        min_lat = min(min_lat, min(netatmo_data[dt]["lat"]))
        min_lon = min(min_lon, min(netatmo_data[dt]["lon"]))
        max_lat = max(max_lat, max(netatmo_data[dt]["lat"]))
        max_lon = max(max_lon, max(netatmo_data[dt]["lon"]))
        minval_dataset_var = min(
            minval_dataset_var, min(netatmo_data[dt][dataset_var])
        )
        maxval_dataset_var = max(
            maxval_dataset_var, max(netatmo_data[dt][dataset_var])
        )

    fig_dict, sliders_dict = init_fig_dict(dataset_var, frame_duration)

    # make frames
    previous_dt = datetime.datetime(2002, 2, 20)
    for dt in netatmo_data:
        if (dt - previous_dt).total_seconds() < min_dt_in_seconds:
            continue
        previous_dt = dt

        frame, slider_step = generate_single_frame(
            netatmo_data[dt], dataset_var, dt, frame_duration
        )

        fig_dict["frames"].append(frame)
        sliders_dict["steps"].append(slider_step)

    fig_dict["layout"]["sliders"] = [sliders_dict]
    fig_dict["data"] = fig_dict["frames"][0]["data"]

    fig = go.Figure(fig_dict)

    return fig


##################################################################################
def empty_scattergeo():
    fig = go.Figure(go.Scattergeo())
    fig.update_layout(
        height=500,
        margin={"r": 0, "t": 0, "l": 0, "b": 0},
        geo=dict(
            # scope='europe',
            showcountries=True,
            # showsubunits = True,
            projection_type="stereographic",
            # center=dict(lon=63, lat=17),
            # projection_rotation=dict(lon=63, lat=17, roll=0),
            lataxis_range=[55, 72],
            lonaxis_range=[4.5, 31.85],
        ),
    )
    return fig


def make_dbscan_fig(df, quiet=True):
    # Knowing whether the scan method was dbscan or hdbscan
    if "is_core_point" in df.columns:
        clustering_method = "dbscan"
    else:
        clustering_method = "hdbscan"
    # Getting member counts for every label
    unique_labels, member_counts = np.unique(
        df["cluster_label"], return_counts=True
    )
    label2count = {
        label: count for label, count in zip(unique_labels, member_counts)
    }

    # Get labels of obs removed due to post-clustering refinement
    # See clustering.py for an explanation about the meaning of
    # rm_obs_label_offset and the convention employed to calculate it
    # and determine these labels.
    try:
        max_abs_rm_label = abs(min(l for l in unique_labels if l < -1))
        rm_obs_label_offset = 1
        while max_abs_rm_label % rm_obs_label_offset != max_abs_rm_label:
            rm_obs_label_offset *= 10
        rm_obs_label_offset //= 10
    except (ValueError):
        rm_obs_label_offset = None

    # Construct array with some descriptive labels for plot
    plot_labels = []
    label_to_legend_label = {}
    for index, row in df.iterrows():
        label = row["cluster_label"]
        if label == -1:
            legend_label = "Rejected obs: "
        elif label < -1:
            orig_cluster_label = int(-(label + rm_obs_label_offset))
            legend_label = "Cluster {}, obs removed due to refining: ".format(
                orig_cluster_label
            )
        else:
            legend_label = "Cluster {}, accepted obs: ".format(int(label))
        legend_label += "{} obs".format(int(label2count[label]))
        plot_labels.append(legend_label)
        label_to_legend_label[label] = legend_label

    # Define colours and symbols
    color_discrete_sequence = [
        px.colors.qualitative.Light24,
        px.colors.qualitative.Alphabet,
        px.colors.qualitative.Dark24,
    ][0]
    # See all symbols at
    # <https://plotly.com/python/marker-style/#custom-marker-symbols>
    symbols = [
        "circle",
        "octagon",
        "hexagon2",
        "circle-dot",
        "octagon-dot",
        "circle-cross",
        "circle-x",
    ]
    # Assign colous and synbols for the legend entries corresponding
    # to every cluster
    color_discrete_map = {-1: "black"}
    leg2symbol = {}
    for ilabel, label in enumerate(l for l in unique_labels if l >= 0):
        # Change colour for every cluster, cycle when all have been used
        color = color_discrete_sequence[ilabel % len(color_discrete_sequence)]
        # Change symbols every len(color_discrete_sequence) time. Cycle when
        # all symbols have been used.
        i_symb = (ilabel // len(color_discrete_sequence)) % len(symbols)
        symbol = symbols[i_symb]
        leg_label = label_to_legend_label[label]
        color_discrete_map[leg_label] = color
        leg2symbol[leg_label] = symbol
        if rm_obs_label_offset is not None:
            try:
                # Make sure points removed by clusters at the refinement stage
                # keep the colours of their parent clusters
                rm_ob_label = int(-(label + rm_obs_label_offset))
                leg_label_of_removed = label_to_legend_label[rm_ob_label]
                color_discrete_map[leg_label_of_removed] = color
                leg2symbol[leg_label_of_removed] = symbol
            except (KeyError):
                if rm_ob_label in unique_labels:
                    raise
                else:
                    # Some clusters do not have any observation removed
                    # after refining
                    pass

    # Preparing plot
    fig = px.scatter_geo(
        df,
        lat="lat",
        lon="lon",
        color=plot_labels,
        color_discrete_map=color_discrete_map,
        hover_name=plot_labels,  # column added to hover information
        hover_data=df.columns,
        projection="stereographic",
        height=500,
    )

    # Zoom in based on coords of accepted obs
    coords_of_accepted_obs = df[df["cluster_label"] > -1][["lat", "lon"]]
    max_lat = max(coords_of_accepted_obs["lat"]) + 0.5
    min_lat = min(coords_of_accepted_obs["lat"]) - 0.5
    max_lon = max(coords_of_accepted_obs["lon"]) + 0.5
    min_lon = min(coords_of_accepted_obs["lon"]) - 0.5

    fig.update_layout(
        clickmode="event+select",
        legend_title=(
            "<b> Clustering of Observations </b> <br>"
            + "<b> Method: {} </b>".format(clustering_method)
        ),
        margin={"r": 0, "t": 0, "l": 0, "b": 0},
        geo=dict(
            showland=True,
            showcountries=True,
            # showsubunits = True,
            lataxis=dict(
                showgrid=True, dtick=10, range=[min_lat - 0.5, max_lat + 0.5],
            ),
            lonaxis=dict(
                showgrid=True, dtick=15, range=[min_lon - 0.5, max_lon + 0.5],
            ),
        ),
    )

    def config_trace_markers(trace):
        tracename_lower = trace.name.lower()
        if "rejected" in tracename_lower:
            trace.update(
                visible="legendonly",
                marker=dict(
                    symbol="x-open",
                    size=5,
                    color="black",
                    opacity=0.25,
                    line=dict(width=1,),
                ),
            )
        elif "removed" in tracename_lower:
            trace.update(
                marker=dict(
                    # Keep colour of original cluster in this case
                    symbol="x-open",
                    size=5,
                    opacity=0.25,
                    line=dict(width=1,),
                )
            )
        else:
            trace.update(
                marker=dict(
                    # symbol="circle",
                    symbol=leg2symbol[trace.name],
                    size=8,
                    opacity=0.75,
                    line=dict(
                        # color=color_discrete_map[trace.name],
                        width=1.0,
                    ),
                )
            )

    fig.for_each_trace(lambda trace: config_trace_markers(trace),)

    return fig



# from .load_data import epoch2datetime
from netatmoqc.load_data import epoch2datetime


def netatmo_csv_to_obsoul(fpath):
    pass


def save_df_as_obsoul(df):
    valid_data_cols = set(
        [
            #'pressure', 'mslp', 'temperature', 'humidity', 'sum_rain_1'
            "temperature",
            "humidity",
            "sum_rain_1",
        ]
    )
    df_data_cols = valid_data_cols.intersection(df.columns)
    # n_params is the same as "number of bodies" in obsoul lingo
    n_params = len(df_data_cols)

    # Construct header.
    # We should have one header for each (id, lat, lon, alt). Since we
    # forbid, in this package, that stations move (lat, lon, alt) within
    # any given DTG, we can say that we should have one header for each id.
    # Moreover, since we are also removing duplicate station IDs within any
    # given DTG, we can say we should have one header for each dataframe row.
    # Values flagged with -1 are values I still don't know
    df["time_utc"] = epoch2datetime(df["time_utc"])
    obs_type = 1  # 1: SYNOP
    obs_code = 14  # 14: Automated land synop
    obs_quality_flag = "1111"  # Got from Jelena
    site_dependent_flag = 100000  # Got from Jelena
    var2varcode = dict(
        # I got these from https://apps.ecmwf.int/odbgov/varno/
        mslp=108,  # varname="pmsl"
        pressure=110,  # varname="ps"
        temperature=39,  # varname="t2m"
        humidity=58,  # varname="rh2m". Or should I use 7 (varname=q)?
        # sum_rain_1 is prob not OK. I got it from
        # https://apps.ecmwf.int/odbgov/accumulationkind/
        sum_rain_1=39001,
    )
    with open("outfile_tmp.txt", "w") as f:
        date = df._parent_dtg.strftime("%Y%m%d")
        hour = df._parent_dtg.strftime("%H")
        f.write("{} {}\n".format(date, hour))
        for row in df.itertuples(index=False):
            header = (
                17,  # Got this from Jelena
                obs_type,
                obs_code,
                row.lat,
                row.lon,
                "'{}'".format(row.id[7:15]),
                date,
                hour,
                row.alt,
                n_params,
                obs_quality_flag,
                site_dependent_flag,
            )

            f.write(" ".join(map(str, header)) + "\n")
            # Construct bodies
            # Values flagged with -1 are values I still don't know
            for param_name in df_data_cols:
                # param_type is the same as "varid@body" in odb lingo
                # See http://www.rclace.eu/File/Data_Assimilation/2007/lace_obspp.pdf:
                # for some explanation on the vertical coordinate conventions for
                # various obs types. In particular, that document states that, for
                # synop and ship, MSLP should, if available, be used as vertical coord
                # instead of pressure, but the MSLP value should be multiplied by -1
                param_type = var2varcode[param_name]
                param_quality_flag = "ne"  # Got from Jelena
                try:
                    first_vert_coord = row.pressure
                except (AttributeError):
                    first_vert_coord = -row.mslp
                # Got the sentinel value 1.7000000000000000E+038 from CARRA OBSOULs
                second_vert_coord = 1.7e038
                body = (
                    param_type,
                    first_vert_coord,
                    second_vert_coord,
                    getattr(row, param_name),
                    param_quality_flag,
                )
                f.write(" ".join(map(str, body)) + "\n")
        print("")


if __name__ == "__main__":
    from netatmoqc.config_parser import ParsedConfig, UndefinedConfigValue
    from netatmoqc.load_data import read_netatmo_data_for_dtg

    config = ParsedConfig(global_default=UndefinedConfigValue)
    df = read_netatmo_data_for_dtg(
        config.dtgs[0], recover_pressure_from_mslp=True, drop_mslp=True
    )
    save_df_as_obsoul(df)



#!/usr/bin/env python3
import logging

logger = logging.getLogger(__name__)


def mpi_parallel(fun, iterable):
    try:
        from mpi4py import MPI
        from mpi4py.futures import MPIPoolExecutor

        # Prevent using mpiexec with n>1
        comm = MPI.COMM_WORLD
        size = comm.Get_size()
        rank = comm.Get_rank()
        if size > 1:
            if rank == 0:
                logger.error(
                    "Received '-n %d' from the MPI runner. Please " % (size)
                    + "use '-n 1' when running this application with MPI, "
                    + "and then select the maximum number N of parallel "
                    + "MPI tasks by passing '-usize N'"
                )
            sys.exit(-1)
    except ImportError as e:
        logger.exception(e)
        logger.error("Support to MPI seems to be unavailable!")
        sys.exit(-1)

    # Establish max #workers that will be dynamically spawn
    # (i.a) Check if universe size passed explicitely (e.g., via "-usize")
    univ_size = comm.Get_attr(MPI.UNIVERSE_SIZE)
    # (i.b) If not (i.a), check if we can guess from the scheduler's env
    if univ_size is None:
        univ_size = os.getenv("SLURM_NTASKS", os.getenv("PBS_NP", None))
    # (i.c) If not (i.b), assume it's a local run
    if univ_size is None:
        univ_size = int(os.getenv("NETATMOQC_MAX_PYTHON_PROCS", 1))
    # (ii) Check how many tasks will actually be performed
    max_workers = min(len(iterable), int(univ_size))

    # Spawn workers and perform tasks
    with MPIPoolExecutor(max_workers=max_workers) as executor:
        results = executor.map(fun, iterable)
    return list(results)
